# linkcheck.tcl - Tcl link checker.
# $Id: linkcheck.tcl,v 1.3 2003/08/04 09:41:34 davidw Exp $
# Copyright(C) 2002, 2003 David N. Welton <davidw@dedasys.com>
#
# See the file license.terms for terms of use and distribution.

package provide linkcheck 0.1

# These are all available from tcllib.

package require http
package require htmlparse
package require struct
package require uri

# These aren't in uri.tcl before 1.1.2.

if { [package vcompare [lindex [package versions uri] end] 1.1.2] < 0 } {

    proc uri::resolve {base url} {
	if {[string length $url]} {
	    if {[isrelative $url]} {

		array set baseparts [split $base]

		switch -- $baseparts(scheme) {
		    http -
		    https -
		    ftp -
		    file {
			array set relparts [split $url]
			if { [string match /* $url] } {
			    catch { set baseparts(path) $relparts(path) }
			} elseif { [string match */ $baseparts(path)] } {
			    set baseparts(path) "$baseparts(path)$relparts(path)"
			} else {
			    if { [string length $relparts(path)] > 0 } {
				set path [lreplace [::split $baseparts(path) /] end end]
				set baseparts(path) "[::join $path /]/$relparts(path)"
			    }
			}
			catch { set baseparts(query) $relparts(query) }
			catch { set baseparts(fragment) $relparts(fragment) }
			return [eval join [array get baseparts]]
		    }
		    default {
			return -code error "unable to resolve relative URL \"$url\""
		    }
		}

	    } else {
		return $url
	    }
	} else {
	    return $base
	}
    }

    proc uri::JoinHttpInner {scheme defport args} {
	array set components [list \
				  host {} port $defport path {} query {} \
				 ]
	array set components $args

	set port {}
	if {[string length $components(port)] && $components(port) != $defport} {
	    set port :$components(port)
	}

	set query {}
	if {[string length $components(query)]} {
	    set query ?$components(query)
	}

	regsub -- {^/} $components(path) {} components(path)

	if { [info exists components(fragment)] && $components(fragment) != "" } {
	    set components(fragment) "#$components(fragment)"
	} else {
	    set components(fragment) ""
	}

	return $scheme://$components(host)$port/$components(path)$components(fragment)$query
    }

    proc uri::canonicalize uri {

	if {[catch {array set u [uri::split $uri]}]} {
	    return $uri
	}
	if {![info exists u(path)]} {
	    return $uri
	}

	set uri $u(path)

	# Remove leading "./" "../" "/.." (and "/../")
	regsub -all -- {^(\./)+}    $uri {}  uri
	regsub -all -- {^/(\.\./)+} $uri {/} uri
	regsub -all -- {^(\.\./)+}  $uri {}  uri

	# Remove inner /./ and /../
	while {[regsub -all -- {/\./}         $uri {/} uri]} {}
	while {[regsub -all -- {/[^/]+/\.\./} $uri {/} uri]} {}
	while {[regsub -all -- {^[^/]+/\.\./} $uri {}  uri]} {}
	# This should cut /.. off the end and go up a directory.
	while {[regsub -all -- {/[^/]+/\.\.} $uri {/} uri]} {}

	set u(path) $uri
	set uri [eval uri::join [array get u]]

	return $uri
    }

}
# -----------------------

namespace eval linkcheck {
    namespace export Init Robots Configure Run Stats End
}

# linkcheck::progresscmd --
#
#	This is the default command used to output some information
#	about the site checking progress.  It can be replaced by
#	something else should one wish to create a different UI.
#
# Arguments:
#	oput - the text to display.
#
# Side Effects:
#	None.
#
# Results:
#	None.

proc linkcheck::progresscmd {oput} {
    puts "$oput"
}

# linkcheck::gethref --
#
#	This command is passed to the HTML parser in order to get the
#	hrefs out of a page.
#
# Arguments:
#	Documented in the 'htmlparlser' man page.
#
# Side Effects:
#	Populates the 'hrefs' list.
#
# Results:
#	None.

proc linkcheck::gethref {tag end attrs contents} {
    variable hrefs
    set tag [string tolower $tag]
    if { $tag == "a" && $end == "" } {
	foreach at [split $attrs] {
	    if { [string match -nocase "href=*" $at] } {
		lappend hrefs [string trim [string range $at 5 end] "\"'"]
	    }
	}
    } elseif { $tag == "img" && $end == "" } {
	foreach at [split $attrs] {
	    if { [string match -nocase "src=*" $at] } {
		lappend hrefs [string trim [string range $at 4 end] "\"'"]
	    }
	}
    } elseif { $tag == "frame" && $end == "" } {
	foreach at [split $attrs] {
	    if { [string match -nocase "src=*" $at] } {
		lappend hrefs [string trim [string range $at 4 end] "\"'"]
	    }
	}
    }
}

# linkcheck::Hrefs --
#
#	Given a block of HTML, and the href of where it came from, add
#	the href's to the list of href's to be processed, and collect
#	some information about them.
#
# Arguments:
#	html - block of HTML.
#	href - where the HTML came from.
#
# Side Effects:
#	Populates the 'from' array with the url's, and the href of the
#	page that contains them.
#
#	Populate the 'spotted' array, so we know what we've already
#	looked at, as well as how deep we are into the links.
#
#	Reset 'hrefs' array so that it doesn't keep filling up from
#	the gethref proc.
#
# Results:
#	None.

proc linkcheck::Hrefs {html href} {
    variable hrefs {}
    variable qu
    variable spotted
    variable from
    variable progresscmd
    if { [catch {
	htmlparse::parse -cmd linkcheck::gethref $html
    } err]} {
	puts $err
	puts $html
	exit 1
    }
    foreach hr $hrefs {
	set uri [uri::resolve $href $hr]
	set orig $uri
	set uri [uri::canonicalize $uri]
	if { [catch {
	    array set attrs [uri::split $uri]
	} err] } {
	    $progresscmd "Unknown url scheme: $hr"
	    continue
	}
	set attrs(fragment) {}
	set uri [eval uri::join [array get attrs]]
	set from($href,$uri) 1
	if { [info exists spotted($uri)] || $attrs(scheme) != "http" } {
	    continue
	}
	$qu put $uri
	set spotted($uri) [expr $spotted($href) + 1]
    }
}

# linkcheck::Page --
#
#	Fetch page local to the web site in question, and analyze the
#	HTML for links.
#
# Arguments:
#	href - the page to get.
#
# Side Effects:
#	Populates 'UrlStats' statistics array with information about
#	the return code generated.
#
# Results:
#	None.

proc linkcheck::Page { href } {
    variable qu
    variable hrefs
    variable UrlStats
    array set attrs [uri::split $href]
    set pg [http::geturl $href]
    set code [http::ncode $pg]
    set UrlStats($href) $code
    upvar 0 $pg state
    array set meta $state(meta)
    set ct {}
    catch { set ct $meta(Content-Type) }
    # In case the server returns it wrong:
    catch { set ct $meta(Content-type) }
    if { $code >= 200 && $code < 400 && [string match "text/html*" $ct] } {
	set data [http::data $pg]
	Hrefs $data $href
    }
    http::cleanup $pg
}

# linkcheck::Link --
#
#	Attempt to fetch 'external' (some other site) link and examine
#	the error code.
#
# Arguments:
#	href - the page to get.
#
# Side Effects:
#       Keeps track of sites that were impossible to fetch because of
# 	networking errors in 'badsites'.  Keeps track of statistics in
# 	'UrlStats' array.
#
# Results:
#	None.

proc linkcheck::Link { href } {
    variable UrlStats
    variable badsites
    array set attrs [uri::split $href]
    if { [info exists badsites($attrs(host))] } {
	set UrlStats($href) 404
	return
    }
    if { [catch {
	set pg [http::geturl $href]
    } err] } {
	set UrlStats($href) 404
	set badsites($attrs(host)) 1
    } else {
	set UrlStats($href) [http::ncode $pg]
	http::cleanup $pg
    }
}

# linkcheck::Init --
#
#	Initializes the link checker.
#
# Arguments:
#	bs - href to use as the base URL for link checking.  Must be a
#	complete URL.
#
# Side Effects:
#	Sets default values for 'delay' (the time to delay between
#	fetching url's), and 'depth' (how far away from the root link
#	to follow links).  A depth of '0' keeps chasing links forever.
#
# Results:
#	None.

proc linkcheck::Init {bs} {
    variable qu
    variable base $bs
    variable delay 10
    variable depth 10
    variable badsites
    array set badsites {}
    variable badpaths {}
    variable progresscmd linkcheck::progresscmd
    set qu [struct::queue]
    $qu put $bs
}

# linkcheck::Robots --
#
#	Fetches the robots.txt file from the site you are checking,
#	and adds the paths described therein to a list of paths not to
#	check.
#
# Arguments:
#	None.
#
# Side Effects:
#	Populates the 'badpaths' list with paths to avoid.
#
# Results:
#	None.

proc linkcheck::Robots {} {
    variable base
    variable badpaths
    variable progresscmd
    array set attrs [uri::split $base]
    set url [uri::join scheme $attrs(scheme) host $attrs(host) port $attrs(port) path robots.txt]
    set pg [http::geturl $url]
    set data [http::data $pg]
    http::cleanup $pg
    foreach ln [split $data \n] {
	if { [regexp "^Disallow:\\s+(.+)" $ln x path] } {
	    lappend badpaths $path
	    $progresscmd "Disabling $path"
	}
    }
}

# linkcheck::Configure --
#
#	Configure options in the linkcheck package.
#
# Arguments:
#	parm - parameter to configure.
#	args - arguments to the parameter.
#
# Side Effects:
#	Change configuration options, which are currently 'delay',
#	'depth', and 'progresscmd'.
#
# Results:
#	None.

proc linkcheck::Configure {parm args} {
    variable delay
    variable depth
    variable progresscmd
    switch $parm {
	delay {
	    set delay [lindex $args 0]
	}
	depth {
	    set depth [lindex $args 0]
	}
	progresscmd {
	    set progresscmd [lindex $args 0]
	}
	default {
	    error "$parm is not a valid parameter"
	}
    }
}

# linkcheck::Cget --
#
#	Get values for configuration options.
#
# Arguments:
#	parm - the parameter whose value you want to fetch.
#
# Side Effects:
#	None.
#
# Results:
#	The value of parameter 'parm'.


proc linkcheck::Cget {parm} {
    variable delay
    variable depth
    variable progresscmd
    switch $parm {
	delay {
	    return $delay
	}
	depth {
	    return $depth
	}
	progresscmd {
	    return $progresscmd
	}
	default {
	    error "$parm is not a valid parameter"
	}
    }
}

# linkcheck::Run --
#
#	Start the link checker in motion.  This is where the work is
#	done.
#
# Arguments:
#	None.
#
# Side Effects:
#	Compile statistics about links.
#
# Results:
#	None.

proc linkcheck::Run {} {
    variable qu
    variable badpaths
    variable base
    variable spotted
    variable delay
    variable depth
    variable progresscmd
    set spotted($base) 1
    while { [$qu size] > 0 } {
	set href [$qu get]
	if { $spotted($href) > $depth && $depth != 0} {
	    continue
	}
	array set baseattrs [uri::split $base]
	array set attrs [uri::split $href]

	set cont 0
	foreach pt $badpaths {
	    if { [string match "${pt}*" "/$attrs(path)"] } {
		$progresscmd "Skipping $href due to robots.txt file"
		set cont 1
		break
	    }
	}
	if { $cont == 1 } { continue }

	after $delay
	if { $attrs(host) == $baseattrs(host) } {
	    $progresscmd "Checking local $href $spotted($href)"
	    Page $href
	} else {
	    $progresscmd "Checking remote $href $spotted($href)"
	    Link $href
	}
    }
}

# linkcheck::CodeMatch --
#
#	Return all the url's whose http code (for instance, 200, 301,
#	404, etc...) matches.
#
# Arguments:
#	match - the string to match against, such as 2??, 404, 30?.
#
# Side Effects:
#	None.
#
# Results:
#	A list of matching url's.

proc linkcheck::CodeMatch {match} {
    variable UrlStats
    set retval {}
    foreach st [array names UrlStats] {
	if { [string match $match $UrlStats($st)] } {
	    lappend retval $st
	}
    }
    return $retval
}

# linkcheck::Stats --
#
#	Return various common statistics about broken links found.
#
# Arguments:
#	cmd - command to get specified statistic.
#	args - arguments to the command.
#
# Side Effects:
#	None.
#
# Results:
#	Statistics regarding links.

proc linkcheck::Stats {cmd args} {
    variable UrlStats
    variable badsites
    variable from

    switch $cmd {
	code {
	    return [linkcheck::CodeMatch [lindex $args 0]]
	}
	visited {
	    return [array size UrlStats]
	}
	good {
	    return [llength [linkcheck::CodeMatch "2??"]]
	}
	bad {
	    return [llength [linkcheck::CodeMatch "4??"]]
	}
	redirect {
	    return [llength [linkcheck::CodeMatch "3??"]]
	}
	broken {
	    return [llength [linkcheck::CodeMatch "5??"]]
	}
	linkedfrom {
	    set retval {}
	    foreach ln [array names from "*,[lindex $args 0]"] {
		lappend retval [lindex [split $ln ,] 0]
	    }
	    return $retval
	}
	default {
	    error "Need a command"
	}
    }
}

# linkcheck::End --
#
#	All good things must come to an end, and so it goes with link
#	checking.  Here is where variables are unset to free memory,
#	statistics are dumped, tents are furled, and so on.
#
# Arguments:
#	None.
#
# Side Effects:
#	Statistics gathered are freed.
#
# Results:
#	None.

proc linkcheck::End {} {
    variable qu
    variable UrlStats
    variable badsites
    variable from
    variable badpaths
    $qu destroy
    array unset UrlStats
    array unset badsites
    array unset from
    unset badpaths
}

